#ifndef hashlib_header
#define hashlib_header

/*@/// doc */
// Hashing of 32 bit integers
//
// (c) 2013..2017 by Jasper L. Neumann
// www.sirrida.de / programming.sirrida.de
// E-Mail: info@sirrida.de
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
// First version: 2013-11
// Last change: 2014-02
//
// Based on the minimal hashing code generator of
// Robert John Jenkins Junior (Bob Jenkins)
// http://burtleburtle.net/bob/hash/perfect.html

// I corrected some errors and tweaked code and parameters to speed it up.
// String handling and special treatment of small switch sets are left out.
// tabb containing only zeros is now optimized away.
// I added reversible and simple hashing as well as 2 new a/b hashing methods.
// Now it is a reusable library.

// Bugs corrected:
// 1. tabh was too small for minimal hashing => AV in apply().
// 2. For 17 keys, slow mode, non-minimal the code generator could choose
//    gen_ab_2 (perfhex.c, hexn(), case 2) and the shift right amount for b
//    could get 32. This does not always work as expected since at least for
//    x86 using 32 bit registers such a shift does *nothing* and the resulting
//    code could trap on accessing invalid tabb elements.
// 3. For slen>131072 alen got an undefined value => AV / heap overflow.
// 4. For fast minimal hashing alen was probably guessed smaller than intended
//    (quirk, no real impact).

/*
Here are some notes from Bob Jenkins' sources:

perfect.c: code to generate code for a hash for perfect hashing.
(c) Bob Jenkins, September 1996, December 1999
You may use this code in any way you wish, and it is free.  No warranty.
I hereby place this in the public domain.
Source is http://burtleburtle.net/bob/c/perfect.c

This generates a minimal perfect hash function.  That means, given a
set of n keys, this determines a hash function that maps each of
those keys into a value in 0..n-1 with no collisions.

The perfect hash function first uses a normal hash function on the key
to determine (a,b) such that the pair (a,b) is distinct for all
keys, then it computes a^scramble[tab[b]] to get the final perfect hash.
tab[] is an array of 1-byte values and scramble[] is a 256-term array of
2-byte or 4-byte values.  If there are n keys, the length of tab[] is a
power of two between n/3 and n.

I found the idea of computing distinct (a,b) values in "Practical minimal
perfect hash functions for large databases", Fox, Heath, Chen, and Daoud,
Communications of the ACM, January 1992.  They found the idea in Chichelli
(CACM Jan 1980).  Beyond that, our methods differ.

The key is hashed to a pair (a,b) where a in 0..*alen*-1 and b in
0..*blen*-1.  A fast hash function determines both a and b
simultaneously.  Any decent hash function is likely to produce
hashes so that (a,b) is distinct for all pairs.  I try the hash
using different values of *salt* until all pairs are distinct.

The final hash is (a XOR scramble[tab[b]]).  *scramble* is a
predetermined mapping of 0..255 into 0..smax-1.  *tab* is an
array that we fill in in such a way as to make the hash perfect.

First we fill in all values of *tab* that are used by more than one
key.  We try all possible values for each position until one works.

This leaves m unmapped keys and m values that something could hash to.
If you treat unmapped keys as lefthand nodes and unused hash values
as righthand nodes, and draw a line connecting each key to each hash
value it could map to, you get a bipartite graph.  We attempt to
find a perfect matching in this graph.  If we succeed, we have
determined a perfect hash for the whole set of keys.

*scramble* is used because (a^tab[i]) clusters keys around *a*.
*/
/*@\\\+00CE*/

#include <stdlib.h>
#include <strings.h>
#include <stdint.h>
#include <string.h>

/*@/// standard stuff */
// Tell the compiler of an unused parameter
#ifdef __GNUC__
  #define UNUSED(x) x __attribute__((unused))
#else
  #define UNUSED(x) x
#endif

#ifndef mycall
#ifdef __GNUC__
  // Define your favourite function options here
  // #define mycall __attribute__((fastcall))
  #define mycall __attribute__((regparm(3)))
#endif
#endif

#ifndef mycall
  #define mycall
#endif

typedef uint8_t t_8u;
typedef uint16_t t_16u;
typedef uint32_t t_32u;
typedef uint64_t t_64u;

typedef int8_t t_8s;
typedef int16_t t_16s;
typedef int32_t t_32s;
typedef int64_t t_64s;

typedef void *t_pointer;
typedef bool t_bool;
typedef char t_char;
typedef int t_int;
typedef float t_float;

#define nil 0
#define assigned(f) ((f)!=nil)
/*@\\\+1259*/

/*@/// interface */
// public:
/*@/// typedef enum tq_hash_method */
typedef enum {
  HM_UNASSIGNED,
  HM_REVERSIBLE,
  HM_SIMPLE_AND,
  HM_SIMPLE_SHR,
  HM_SIMPLE_ROL,
  HM_SIMPLE_ROL_XOR,
  HM_SIMPLE_ROL_ADD,
  HM_SIMPLE_ROL_SUB,
  HM_SIMPLE_MUL,
  HM_AB_1,
  HM_AB_2,
  HM_AB_3,
  HM_AB_X1,    // experimental
  HM_AB_X2,    // experimental
  HM_AB_X3,    // experimental
  HM_MAX       // dummy
  } tq_hash_method;
/*@\\\0000000212+5AAE*/
/*@/// typedef struct tr_gen_opt */
typedef struct tr_gen_opt {
  // parameters for the generating callback functions, see tr_callback
  // all values should be treated read-only

  // common parameters
  t_32u a_shl;
  t_32u a_shr;
  t_32u a_mask;  // and factor
  t_32u b_shl;
  t_32u b_shr;
  t_32u b_mask;  // and factor
  t_32u salt;  // and extra stuff

  // for tabular methods
  t_32u *p_scramble;  // array; pointer to the scramble table
  t_32u scramble_len;  // 0=not used
  t_32u scramble_max;
  t_32u *p_tabb;  // array; pointer to the B table
  t_32u tabb_len;
  t_32u tabb_max;
  t_bool do_xor_shr_16;  // extra scrambling
  t_bool do_add_shl_8;  // extra scrambling

  // info
  t_bool check_range;  // check range
  t_bool check_value;  // check if value OK
    // if both are set: reversible hashing; test by bit array or jump table
    // i.e. reversible hashing is not minimal / jump table contains holes
  t_32u hash_used_max;  // maximum used hash value
  t_32u hash_gen_max;  // maximum generated hash value

  // statistics
  t_32u count_for_limit;  // used total work, see calc_opt.work_limit
  t_32u max_count;  // max. work per element, see
  t_32u low;  // lowest key; use as hint to generate value table
  t_32u high;  // highest key
  t_32u bits;  // needed bits for keys
  t_bool use_signed;  // use signed values for keys
  t_bool is_minimal;  // no default entries in jump table
  tq_hash_method hash_method;  // chosen hash method
  } tr_gen_opt;
/*@\\\+A55F*/
typedef void (*tf_generate) (const tr_gen_opt &gen_opt, t_pointer gen_stuff);
typedef void (*tf_step_key) (t_32u key, t_pointer key_stuff, t_pointer step_stuff);

/*@/// typedef struct tr_callback */
typedef struct tr_callback {
  // all values are read/write
  tf_generate a_gen[HM_MAX];  // indexed by tq_hash_method
#if 0
  /*@/// tf_generate gen_reversible; */
  tf_generate gen_reversible;

  /*
  // This is the only hash method here which does not detect dupes
  // OPTIMIZE for gen_opt.a_shr=0 and gen_opt.salt=1

  t_32u calc_rev(t_32u x) {

    return ror_32(x-gen_opt.salt, gen_opt.a_shr) * gen_opt.a_mask;
    }
  */
  /*@\\\+57D3*/

  /*@/// tf_generate gen_simple_and; */
  tf_generate gen_simple_and;

  /*
  t_32u calc_and(t_32u x) {

    return x & gen_opt.a_mask;
    }
  */
  /*@\\\0000000401+1224*/
  /*@/// tf_generate gen_simple_shr; */
  tf_generate gen_simple_shr;

  /*
  // gen_opt.a_shr: 1..31

  t_32u calc_shr(t_32u x) {

    return x >> gen_opt.a_shr;
    }
  */
  /*@\\\0000000501+CAFE*/
  /*@/// tf_generate gen_simple_rol; */
  tf_generate gen_simple_rol;

  /*
  // gen_opt.salt: 1..31

  t_32u calc_rol(t_32u x) {

    return rol_32(x, gen_opt.salt) & gen_opt.a_mask;
    }
  */
  /*@\\\000000080E+44C6*/
  /*@/// tf_generate gen_simple_rol_xor; */
  tf_generate gen_simple_rol_xor;

  /*
  // gen_opt.salt: 1..31

  t_32u calc_rol_xor(t_32u x) {

    return (rol_32(x, gen_opt.salt) ^ x) & gen_opt.a_mask;
    }
  */
  /*@\\\0000000811+0C3E*/
  /*@/// tf_generate gen_simple_rol_add; */
  tf_generate gen_simple_rol_add;

  /*
  // gen_opt.salt: 1..31

  t_32u calc_rol_add(t_32u x) {

    return (rol_32(x, gen_opt.salt) + x) & gen_opt.a_mask;
    }
  */
  /*@\\\0000000822+AFA6*/
  /*@/// tf_generate gen_simple_rol_sub; */
  tf_generate gen_simple_rol_sub;

  /*
  // gen_opt.salt: 1..31

  t_32u calc_rol_sub(t_32u x) {

    return (rol_32(x, gen_opt.salt) - x) & gen_opt.a_mask;
    }
  */
  /*@\\\0000000822+78F7*/
  /*@/// tf_generate gen_simple_mul; */
  tf_generate gen_simple_mul;

  /*
  t_32u calc_mul(t_32u x) {

    return (t_32u)(x * gen_opt.salt) >> gen_opt.a_shr;
    }
  */
  /*@\\\0000000401+D40F*/

  /*@/// tf_generate gen_ab_1; */
  tf_generate gen_ab_1;

  /*
  // OPTIMIZE for gen_opt.a_shl=gen_opt.a_shr
  // gen_opt.a_shr: 1..31

  t_32u calc_ab_1(t_32u x) {

    t_32u a,b;

    a = (x << gen_opt.a_shl) >> gen_opt.a_shr;
    b = (x >> gen_opt.b_shr) & gen_opt.b_mask;
    return calc_tab(a,b);
    }
  */
  /*@\\\+797B*/
  /*@/// tf_generate gen_ab_2; */
  tf_generate gen_ab_2;

  /*
  // OPTIMIZE for gen_opt.b_shl=gen_opt.b_shr
  // gen_opt.a_shr: 1..31

  t_32u calc_ab_2(t_32u x) {

    t_32u a,b;

    a = (x >> gen_opt.a_shr) & gen_opt.a_mask;
    b = (x << gen_opt.b_shl) >> gen_opt.b_shr;
    return calc_tab(a,b);
    }
  */
  /*@\\\+8BA4*/
  /*@/// tf_generate gen_ab_3; */
  tf_generate gen_ab_3;

  /*
  t_32u calc_ab_3(t_32u x) {

    t_32u a,b;

    x += gen_opt.salt;

    if (gen_opt.do_xor_shr_16)
      x ^= (x >> 16);

    if (gen_opt.do_add_shl_8)
      x += (x << 8);

    x ^= (x >> 4);

    if (gen_opt.a_shl == 0)
      a = x >> gen_opt.a_shr;
    else
      a = ((x << gen_opt.a_shl) + x) >> gen_opt.a_shr;

    b = (x >> gen_opt.b_shr) & gen_opt.b_mask;

    return calc_tab(a,b);
    }
  */
  /*@\\\+9DCD*/
#endif
  } tr_callback;

/*
t_32u calc_tab(t_32u a, t_32u b) {

  if (gen_opt.tabb_len == 0)
    return a;  // OPTIMIZE: do not calculate b
  else if (gen_opt.tabb_len == 1)
    return a ^ gen_opt.p_tabb[0];
  // OPTIMIZE: do not calculate b, tabb[0]=const!
  else if (gen_opt.scramble_len == 0)
    return a ^ gen_opt.p_tabb[b];
  else
    return a ^ gen_opt.p_scramble[gen_opt.p_tabb[b]];
  }
*/
/*@\\\0000002201+5678*/
/*@/// typedef struct tr_calc_opt */
typedef struct tr_calc_opt {
  // all values are read/write
  t_32u nr_mul_checks;  // number of tries for gen_simple_mul
  t_32u listlen_limit;  // max. number of elements with same b value (speedup)
  t_32u retry_initkey;  // number of times to try to find distinct (a,b)
  t_32u retry_perfect;  // number of times to try to make a perfect hash
  t_32u use_scramble;  // use scramble if blen >= use_scramble
  t_32u count_limit;  // max. work per element (speedup)
  t_32u work_limit;  // max. work total (speedup), 0=infinity
  t_float min_load_factor;  // <= 1 (better <= 0.5)
  t_float near_minimal_factor;  // >= 1, for ab hashing to limit hash value
  t_float keyspace_factor;  // <= 1, twice key space if
    // self.priv.nkeys > self.slen*self.calc_opt.keyspace_factor, fast: 0.8
  t_bool minimal;  // for minimal hashing, i.e. no holes
  t_bool fast;  // speedup this library at the potential cost of larger tables
  t_bool fast_trans;  // allow transitive closure for fast mode
  } tr_calc_opt;
/*@\\\000000100A+15B0*/

// private:
/*@/// typedef struct tr_key */
typedef struct tr_key *tpr_key;
typedef struct tr_key {
  // representation of a key
  t_32u key_k;  // the initial hash value for this key, aka hash_k
  tpr_key next_k;  // next key

  t_pointer user_k;  // for key only, optional

  // beyond this point is mapping-dependent
  t_32u a_k;  // a, of the key maps to (a,b)
  t_32u b_k;  // b, of the key maps to (a,b)
  tpr_key nextb_k;  // next key with this b
  } tr_key;
/*@\\\+6DA4*/
/*@/// typedef struct tr_bstuff */
typedef struct tr_bstuff *tpr_bstuff;
typedef struct tr_bstuff {
  // things indexed by b of original (a,b) pair
  t_32u val_b;  // hash=a^tabb[b].val_b
  tpr_key list_b;  // tabb[i].list_b is list of keys with b==i
  t_32u listlen_b;  // length of list_b
  t_32u water_b;  // high watermark of who has visited this map node
  } tr_bstuff;
/*@\\\000000080F+B4A5*/
/*@/// typedef struct tr_hstuff */
typedef struct tr_hstuff {
  // things indexed by final hash value
  tpr_key key_h;  // tabh[i].key_h is the key with a hash of i
  } tr_hstuff;
/*@\\\000000040F+445B*/
/*@/// typedef struct tr_qstuff */
typedef struct tr_qstuff *tpr_qstuff;
typedef struct tr_qstuff {
  // things indexed by queue position
  tpr_bstuff b_q;  // b that currently occupies this hash
  t_32u parent_q;  // queue position of parent that could use this hash
  t_32u newval_q;  // what to change parent tab[b] to to use this hash
  t_32u oldval_q;  // original value of tab[b]
  } tr_qstuff;
/*@\\\+1BEC*/
typedef mycall void (*tf_calc_ab) (const tr_gen_opt &opt, tr_key &key);
typedef mycall t_32u (*tf_calc_hash) (const tr_gen_opt &opt, const tr_key &key);
/*@/// typedef struct tr_private */
typedef struct tr_private {
  t_32u nkeys;  // number of keys
  tpr_key keys;  // head of list of keys

  // local tables
  tpr_bstuff p_tabb;  // array; table indexed by b
  t_32u slen;  // p_scramble[] values in 0..slen-1, a power of 2, aka smax
  t_32u alen;  // a in 0..alen-1, a power of 2
  t_32u blen;  // b in 0..blen-1, a power of 2
  t_32u *p_scramble;  // array; used in final hash function

  t_32u lowbit;  // for inithex(), lowest interesting bit
  t_32u highbit;  // for inithex(), highest interesting bit
  t_32u diffbits;  // bits which differ for some key

  // code generator, hash generators
  tr_hstuff *p_inv_hash;  // array; inverse hash
  t_32u allocated_scramble;
  t_32u highhash;  // highest allowed hash value
  t_bool trans;  // do transitive closure
  } tr_private;
/*@\\\+0A6E*/
/*@/// typedef struct tr_state_engine */
typedef struct tr_state_engine {
  // state machine used in inithex()
  t_32u i,j,k;
  } tr_state_engine;
/*@\\\+2819*/
/*@/// typedef struct tr_reversible */
typedef struct tr_reversible {
  // internal transfer record
  t_32u gcd,ror,inv,max;
  } tr_reversible;
/*@\\\+3696*/
/*@/// typedef enum tq_reason */
typedef enum {
  REASON_SUCCESS,  // found distinct (a,b) for all keys, put keys in tabb[]
  REASON_FAILURE,  // didn't find distinct (a,b) for all keys
  REASON_LIMIT,  // limit reached, early abort
  REASON_DUPE  // a real dupe found, cannot proceed
  } tq_reason;
/*@\\\+DCDA*/

// public:
/*@/// class to_perfect_hash */
class to_perfect_hash {

// Fields are ordered such that alignment costs minimal space
// and often used fields need short offsets
private:
  tr_private priv;
  tr_state_engine state_engine;

public:
  tr_calc_opt calc_opt;
  tr_gen_opt gen_opt;
  tr_callback cb;
  t_int ref_count;  // not used by library; can be used by application

public:
  to_perfect_hash () { init(); };
  ~to_perfect_hash () { stop(); };
private:
  // don't clone me
  to_perfect_hash (const to_perfect_hash &) {};
  to_perfect_hash &operator =(const to_perfect_hash&) {return *this;};
public:
  void defaults();
  t_bool self_test() const;  // OK?

  void clean();
  void add_key(t_32u value, t_pointer user_k);
  t_bool generate_hash();
  void do_generate(t_pointer gen_stuff);
  t_32u calc_hash(t_32u key_k) const;
  void step_keys(tf_step_key f_step_key, t_pointer step_stuff) const;
  t_bool inv_hash(t_32u hash, t_32u &key, t_pointer &key_stuff) const;

private:
  void init();
  void stop();
  mycall void calc_max();

  // main hash methods
  mycall t_bool test_reversible();
    mycall void calc_rev_cfg(tr_reversible &rev, t_32u min);
  mycall t_bool test_simple_perfect();
    mycall t_bool check_direct(tq_hash_method hash_method, t_32u x);
  mycall t_bool findhash();
    mycall void initalen();
    mycall void scrambleinit();
    mycall void setlow();
    mycall void inithex(t_32u salt);
    mycall tq_reason inittab();
    /*@/// function apply(tabh,tabq,tail,rollback):t_bool; */
    mycall t_bool apply(
      tr_hstuff *tabh,  // tar_hstuff
      tr_qstuff *tabq,  // tar_qstuff
      t_32u tail,
      t_bool rollback
      );
    /*@\\\000000020E+5EC8*/
    /*@/// function augment(tabh,tabq,item,highwater):tq_reason; */
    mycall tq_reason augment(
      tr_hstuff *tabh,  // tar_hstuff
      tr_qstuff *tabq,  // tar_qstuff
      tpr_bstuff item,
      t_32u highwater
      );
    /*@\\\000000040D+EC2E*/
    /*@/// function perfect(tabh,tabq):tq_reason; */
    mycall tq_reason perfect(
      tr_hstuff *tabh,  // tar_hstuff
      tr_qstuff *tabq  // tar_qstuff
      );
    /*@\\\+FF44*/
    mycall void finalize_perfect();
  };
/*@\\\+0E34*/
/*@\\\0000000801+3181*/

#endif
/*@\\\0000000D01+E94C*/
